@@ -1,63 +1,16 @@
-/* crypto/engine/hw_pkcs11.c */
-/* Written by Geoff Thorpe (geoff@geoffthorpe.net) for the OpenSSL
+/*
+ * Copyright 1999-2001 The OpenSSL Project Authors. All Rights Reserved.
+ * Written by Geoff Thorpe (geoff@geoffthorpe.net) for the OpenSSL
  * project 2000.
+ * Portions Copyright (c) 2003 Kevin Stefanik (kstef@mtppi.org)
  * Copied/modified by Kevin Stefanik (kstef@mtppi.org) for the OpenSC
  * project 2003.
- * Copyright (c) 2017 Michał Trojnara
- */
-/* ====================================================================
- * Copyright (c) 1999-2001 The OpenSSL Project.  All rights reserved.
- * Portions Copyright (c) 2003 Kevin Stefanik (kstef@mtppi.org)
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- *
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- *
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in
- *    the documentation and/or other materials provided with the
- *    distribution.
- *
- * 3. All advertising materials mentioning features or use of this
- *    software must display the following acknowledgment:
- *    "This product includes software developed by the OpenSSL Project
- *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
- *
- * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
- *    endorse or promote products derived from this software without
- *    prior written permission. For written permission, please contact
- *    licensing@OpenSSL.org.
- *
- * 5. Products derived from this software may not be called "OpenSSL"
- *    nor may "OpenSSL" appear in their names without prior written
- *    permission of the OpenSSL Project.
- *
- * 6. Redistributions of any form whatsoever must retain the following
- *    acknowledgment:
- *    "This product includes software developed by the OpenSSL Project
- *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
- *
- * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
- * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
- * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
- * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
- * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
- * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
- * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
- * OF THE POSSIBILITY OF SUCH DAMAGE.
- * ====================================================================
- *
- * This product includes cryptographic software written by Eric Young
- * (eay@cryptsoft.com).  This product includes software written by Tim
- * Hudson (tjh@cryptsoft.com).
+ * Copyright (c) 2016-2018 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
+ * Licensed under the OpenSSL license (the "License").  You may not use
+ * this file except in compliance with the License.  You can obtain a copy
+ * in the file LICENSE in the source distribution or at
+ * https://www.openssl.org/source/license.html
  */
 
 #include "engine.h"
@@ -118,6 +71,10 @@ static const ENGINE_CMD_DEFN engine_cmd_defns[] = {
 		"SET_CALLBACK_DATA",
 		"Set the global user interface extra data (internal)",
 		ENGINE_CMD_FLAG_INTERNAL},
+	{CMD_FORCE_LOGIN,
+		"FORCE_LOGIN",
+		"Force login to the PKCS#11 module",
+		ENGINE_CMD_FLAG_NO_INPUT},
 	{0, NULL, NULL, 0}
 };
 
@@ -144,13 +101,26 @@ static ENGINE_CTX *get_ctx(ENGINE *engine)
 static int engine_destroy(ENGINE *engine)
 {
 	ENGINE_CTX *ctx;
-	int rv;
+	int rv = 1;
 
 	ctx = get_ctx(engine);
 	if (ctx == NULL)
 		return 0;
-	rv = ctx_destroy(ctx);
+
+	/* ENGINE_remove() invokes our engine_destroy() function with
+	 * CRYPTO_LOCK_ENGINE / global_engine_lock acquired.
+	 * Any attempt to re-acquire the lock either by directly
+	 * invoking OpenSSL functions, or indirectly via PKCS#11 modules
+	 * that use OpenSSL engines, causes a deadlock. */
+	/* Our workaround is to skip ctx_finish() entirely, as a memory
+	 * leak is better than a deadlock. */
+#if 0
+	rv &= ctx_finish(ctx);
+#endif
+
+	rv &= ctx_destroy(ctx);
 	ENGINE_set_ex_data(engine, pkcs11_idx, NULL);
+	ERR_unload_ENG_strings();
 	return rv;
 }
 
@@ -168,11 +138,28 @@ static int engine_init(ENGINE *engine)
 static int engine_finish(ENGINE *engine)
 {
 	ENGINE_CTX *ctx;
+	int rv = 1;
 
 	ctx = get_ctx(engine);
 	if (ctx == NULL)
 		return 0;
-	return ctx_finish(ctx);
+
+	/* ENGINE_cleanup() used by OpenSSL versions before 1.1.0 invokes
+	 * our engine_finish() function with CRYPTO_LOCK_ENGINE acquired.
+	 * Any attempt to re-acquire CRYPTO_LOCK_ENGINE either by directly
+	 * invoking OpenSSL functions, or indirectly via PKCS#11 modules
+	 * that use OpenSSL engines, causes a deadlock. */
+	/* Our workaround is to skip ctx_finish() for the affected OpenSSL
+	 * versions, as a memory leak is better than a deadlock. */
+	/* We cannot simply temporarily release CRYPTO_LOCK_ENGINE here, as
+	 * engine_finish() is also executed from ENGINE_finish() without
+	 * acquired CRYPTO_LOCK_ENGINE, and there is no way with to check
+	 * whether a lock is already acquired with OpenSSL < 1.1.0 API. */
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
+	rv &= ctx_finish(ctx);
+#endif
+
+	return rv;
 }
 
 static EVP_PKEY *load_pubkey(ENGINE *engine, const char *s_key_id,
@@ -190,11 +177,21 @@ static EVP_PKEY *load_privkey(ENGINE *engine, const char *s_key_id,
 		UI_METHOD *ui_method, void *callback_data)
 {
 	ENGINE_CTX *ctx;
+	EVP_PKEY *pkey;
 
 	ctx = get_ctx(engine);
 	if (ctx == NULL)
 		return 0;
-	return ctx_load_privkey(ctx, s_key_id, ui_method, callback_data);
+	pkey = ctx_load_privkey(ctx, s_key_id, ui_method, callback_data);
+#ifdef EVP_F_EVP_PKEY_SET1_ENGINE
+	/* EVP_PKEY_set1_engine() is required for OpenSSL 1.1.x,
+	 * but otherwise setting pkey->engine breaks OpenSSL 1.0.2 */
+	if (pkey && !EVP_PKEY_set1_engine(pkey, engine)) {
+		EVP_PKEY_free(pkey);
+		pkey = NULL;
+	}
+#endif /* EVP_F_EVP_PKEY_SET1_ENGINE */
+	return pkey;
 }
 
 static int engine_ctrl(ENGINE *engine, int cmd, long i, void *p, void (*f) ())
@@ -234,10 +231,12 @@ static int bind_helper(ENGINE *e)
 			!ENGINE_set_ECDH(e, PKCS11_get_ecdh_method()) ||
 #endif
 #endif /* OPENSSL_VERSION_NUMBER */
+			!ENGINE_set_pkey_meths(e, PKCS11_pkey_meths) ||
 			!ENGINE_set_load_pubkey_function(e, load_pubkey) ||
 			!ENGINE_set_load_privkey_function(e, load_privkey)) {
 		return 0;
 	} else {
+		ERR_load_ENG_strings();
 		return 1;
 	}
 }
