@@ -29,6 +29,7 @@
 
 #ifndef OPENSSL_NO_EC
 #include <openssl/ec.h>
+#include <openssl/bn.h>
 #endif
 #ifndef OPENSSL_NO_ECDSA
 #include <openssl/ecdsa.h>
@@ -204,7 +205,7 @@ static int pkcs11_get_params(EC_KEY *ec, PKCS11_KEY *key)
 
 /* Retrieve EC point from key into ec
  * return nonzero on error */
-static int pkcs11_get_point(EC_KEY *ec, PKCS11_KEY *key)
+static int pkcs11_get_point_key(EC_KEY *ec, PKCS11_KEY *key)
 {
 	CK_BYTE *point;
 	size_t point_len = 0;
@@ -232,6 +233,40 @@ static int pkcs11_get_point(EC_KEY *ec, PKCS11_KEY *key)
 	return rv;
 }
 
+/* Retrieve EC point from cert into ec
+ * return nonzero on error */
+static int pkcs11_get_point_cert(EC_KEY *ec, PKCS11_CERT *cert)
+{
+	EVP_PKEY *pubkey = NULL;
+	EC_KEY *pubkey_ec;
+	const EC_POINT *point;
+	int rv = -1;
+
+	if (cert == NULL)
+		goto error;
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
+	pubkey = X509_get0_pubkey(cert->x509);
+#else
+	pubkey = X509_get_pubkey(cert->x509);
+#endif
+	if (pubkey == NULL)
+		goto error;
+	pubkey_ec = EVP_PKEY_get0_EC_KEY(pubkey);
+	if (pubkey_ec == NULL)
+		goto error;
+	point = EC_KEY_get0_public_key(pubkey_ec);
+	if (point == NULL)
+		goto error;
+	if (EC_KEY_set_public_key(ec, point) == 0)
+		goto error;
+	rv = 0;
+error:
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+	EVP_PKEY_free(pubkey);
+#endif
+	return rv;
+}
+
 static EC_KEY *pkcs11_get_ec(PKCS11_KEY *key)
 {
 	EC_KEY *ec;
@@ -247,9 +282,17 @@ static EC_KEY *pkcs11_get_ec(PKCS11_KEY *key)
 	 * it if the PKCS#11 module or the hardware can figure this out
 	 */
 	no_params = pkcs11_get_params(ec, key);
-	no_point = pkcs11_get_point(ec, key);
+	no_point = pkcs11_get_point_key(ec, key);
 	if (no_point && key->isPrivate) /* Retry with the public key */
-		no_point = pkcs11_get_point(ec, pkcs11_find_key_from_key(key));
+		no_point = pkcs11_get_point_key(ec, pkcs11_find_key_from_key(key));
+	if (no_point && key->isPrivate) /* Retry with the certificate */
+		no_point = pkcs11_get_point_cert(ec, pkcs11_find_certificate(key));
+
+	if (key->isPrivate && EC_KEY_get0_private_key(ec) == NULL) {
+		BIGNUM *bn = BN_new();
+		EC_KEY_set_private_key(ec, bn);
+		BN_free(bn);
+	}
 
 	/* A public keys requires both the params and the point to be present */
 	if (!key->isPrivate && (no_params || no_point)) {
@@ -260,7 +303,7 @@ static EC_KEY *pkcs11_get_ec(PKCS11_KEY *key)
 	return ec;
 }
 
-static PKCS11_KEY *pkcs11_get_ex_data_ec(const EC_KEY *ec)
+PKCS11_KEY *pkcs11_get_ex_data_ec(const EC_KEY *ec)
 {
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
 	return EC_KEY_get_ex_data(ec, ec_ex_index);
