@@ -51,21 +51,19 @@ int pkcs11_set_ui_method(PKCS11_CTX *ctx,
 }
 
 /*
- * Find key matching a certificate
+ * Find private key matching a certificate
  */
 PKCS11_KEY *pkcs11_find_key(PKCS11_CERT *cert)
 {
-	PKCS11_CERT_private *cpriv;
-	PKCS11_KEY_private *kpriv;
+	PKCS11_CERT_private *cpriv = PRIVCERT(cert);
 	PKCS11_KEY *keys;
 	unsigned int n, count;
 
-	cpriv = PRIVCERT(cert);
-	if (PKCS11_enumerate_keys(CERT2TOKEN(cert), &keys, &count))
+	if (pkcs11_enumerate_keys(CERT2TOKEN(cert), CKO_PRIVATE_KEY, &keys, &count))
 		return NULL;
 	for (n = 0; n < count; n++) {
-		kpriv = PRIVKEY(&keys[n]);
-		if (cpriv->id_len == kpriv->id_len
+		PKCS11_KEY_private *kpriv = PRIVKEY(&keys[n]);
+		if (kpriv && cpriv->id_len == kpriv->id_len
 				&& !memcmp(cpriv->id, kpriv->id, cpriv->id_len))
 			return &keys[n];
 	}
@@ -79,14 +77,14 @@ PKCS11_KEY *pkcs11_find_key_from_key(PKCS11_KEY *keyin)
 {
 	PKCS11_KEY_private *kinpriv = PRIVKEY(keyin);
 	PKCS11_KEY *keys;
-	unsigned int n, count;
+	unsigned int n, count, type =
+		keyin->isPrivate ? CKO_PUBLIC_KEY : CKO_PRIVATE_KEY;
 
-	pkcs11_enumerate_keys(KEY2TOKEN(keyin),
-		keyin->isPrivate ? CKO_PUBLIC_KEY : CKO_PRIVATE_KEY, /* other type */
-		&keys, &count);
+	if (pkcs11_enumerate_keys(KEY2TOKEN(keyin), type, &keys, &count))
+		return NULL;
 	for (n = 0; n < count; n++) {
 		PKCS11_KEY_private *kpriv = PRIVKEY(&keys[n]);
-		if (kinpriv->id_len == kpriv->id_len
+		if (kpriv && kinpriv->id_len == kpriv->id_len
 				&& !memcmp(kinpriv->id, kpriv->id, kinpriv->id_len))
 			return &keys[n];
 	}
@@ -331,10 +329,11 @@ EVP_PKEY *pkcs11_get_key(PKCS11_KEY *key, int isPrivate)
 		if (key->evp_key == NULL)
 			return NULL;
 		kpriv->always_authenticate = CK_FALSE;
-		if(isPrivate) {
-			if(key_getattr_val(key, CKA_ALWAYS_AUTHENTICATE,
-					&kpriv->always_authenticate, sizeof(CK_BBOOL)))
-				fprintf(stderr, "Missing CKA_ALWAYS_AUTHENTICATE attribute\n");
+		if (isPrivate && key_getattr_val(key, CKA_ALWAYS_AUTHENTICATE,
+				&kpriv->always_authenticate, sizeof(CK_BBOOL))) {
+#ifdef DEBUG
+			fprintf(stderr, "Missing CKA_ALWAYS_AUTHENTICATE attribute\n");
+#endif
 		}
 	}
 #if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
@@ -448,7 +447,7 @@ int pkcs11_enumerate_keys(PKCS11_TOKEN *token, unsigned int type,
 
 /**
  * Remove a key from the associated token
- */ 
+ */
 int pkcs11_remove_key(PKCS11_KEY *key) {
 	PKCS11_SLOT *slot = KEY2SLOT(key);
 	PKCS11_CTX *ctx = KEY2CTX(key);
@@ -457,6 +456,7 @@ int pkcs11_remove_key(PKCS11_KEY *key) {
 	CK_ULONG count;
 	CK_ATTRIBUTE search_parameters[32];
 	unsigned int n = 0;
+	int rv;
 
 	/* First, make sure we have a session */
 	if (!spriv->haveSession && PKCS11_open_session(slot, 1))
@@ -470,7 +470,7 @@ int pkcs11_remove_key(PKCS11_KEY *key) {
 	if (key->label)
 	 	pkcs11_addattr_s(search_parameters + n++, CKA_LABEL, key->label);
 
-	int rv = CRYPTOKI_call(ctx,
+	rv = CRYPTOKI_call(ctx,
 		C_FindObjectsInit(spriv->session, search_parameters, n));
 	CRYPTOKI_checkerr(CKR_F_PKCS11_REMOVE_KEY, rv);
 
