@@ -1,5 +1,6 @@
 /* libp11, a simple layer on to of PKCS#11 API
  * Copyright (C) 2005 Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2015 Michał Trojnara <Michal.Trojnara@stunnel.org>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -16,19 +17,9 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
  */
 
-#include <config.h>
+#include "libp11-int.h"
 #include <string.h>
 #include <openssl/crypto.h>
-#include "libp11-int.h"
-
-void *pkcs11_malloc(size_t size)
-{
-	void *p = OPENSSL_malloc(size);
-	if (p == NULL)
-		return NULL;
-	memset(p, 0, size);
-	return p;
-}
 
 /* PKCS11 strings are fixed size blank padded,
  * so when strduping them we must make sure
@@ -40,7 +31,7 @@ char *pkcs11_strdup(char *mem, size_t size)
 
 	while (size && mem[size - 1] == ' ')
 		size--;
-	res = (char *) OPENSSL_malloc(size + 1);
+	res = OPENSSL_malloc(size + 1);
 	if (res == NULL)
 		return NULL;
 	memcpy(res, mem, size);
@@ -49,15 +40,31 @@ char *pkcs11_strdup(char *mem, size_t size)
 }
 
 /*
- * Dup memory
+ * CRYPTO dynlock wrappers: 0 is an invalid dynamic lock ID
  */
-void *memdup(const void *src, size_t size)
-{
-	void *dst;
 
-	dst = malloc(size);
-	if (dst == NULL)
-		return NULL;
-	memcpy(dst, src, size);
-	return dst;
+#if OPENSSL_VERSION_NUMBER < 0x10100004L || defined(LIBRESSL_VERSION_NUMBER)
+
+int CRYPTO_THREAD_lock_new()
+{
+	int i;
+
+	if (CRYPTO_get_dynlock_create_callback() == NULL ||
+			CRYPTO_get_dynlock_lock_callback() == NULL ||
+			CRYPTO_get_dynlock_destroy_callback() == NULL)
+		return 0; /* Dynamic callbacks not set */
+	i = CRYPTO_get_new_dynlockid();
+	if (i == 0)
+		ERR_clear_error(); /* Dynamic locks are optional -> ignore */
+	return i;
 }
+
+void CRYPTO_THREAD_lock_free(int i)
+{
+	if (i)
+		CRYPTO_destroy_dynlockid(i);
+}
+
+#endif
+
+/* vim: set noexpandtab: */
